#!/usr/bin/env python3

import zmq
import logging

logger = logging.getLogger(__name__)


class TasksOutgoing(object):
    """ TODO: docstring """

    def __init__(self, ip_address, port_range):
        """ TODO: docstring """
        self.context = zmq.Context()
        self.zmq_socket = self.context.socket(zmq.DEALER)
        self.zmq_socket.set_hwm(0)
        self.port = self.zmq_socket.bind_to_random_port("tcp://{}".format(ip_address),
                                                        min_port=port_range[0],
                                                        max_port=port_range[1])
        self.poller = zmq.Poller()
        self.poller.register(self.zmq_socket, zmq.POLLOUT)

    def put(self, task_id, buffer):
        """ TODO: docstring """
        task_id_bytes = task_id.to_bytes(4, "little")
        message = [b"", task_id_bytes] + buffer

        self.zmq_socket.send_multipart(message)
        logger.debug("Sent task {}".format(task_id))

    def close(self):
        self.zmq_socket.close()
        self.context.term()


class ResultsIncoming(object):
    """ TODO: docstring """

    def __init__(self, ip_address, port_range):
        """ TODO: docstring """
        self.context = zmq.Context()
        self.zmq_socket = self.context.socket(zmq.DEALER)
        self.zmq_socket.set_hwm(0)
        self.port = self.zmq_socket.bind_to_random_port(
            "tcp://{}".format(ip_address),
            min_port=port_range[0],
            max_port=port_range[1])

    def get(self):

        result = self.zmq_socket.recv_multipart()
        task_id = int.from_bytes(result[1], "little")
        buffer = result[2:]
        return task_id, buffer

    def close(self):
        self.zmq_socket.close()
        self.context.term()


class WorkerMessages(object):
    """ TODO: docstring """

    def __init__(self, tasks_url):
        self.context = zmq.Context()
        self.zmq_socket = self.context.socket(zmq.REP)
        self.zmq_socket.connect(tasks_url)

    def get(self):
        bufs = self.zmq_socket.recv_multipart()
        task_id = int.from_bytes(bufs[0], "little")
        return task_id, bufs[1:]

    def put(self, task_id, buffer):
        task_id_bytes = task_id.to_bytes(4, "little")
        self.zmq_socket.send_multipart([task_id_bytes] + buffer)

    def close(self):
        self.zmq_socket.close()
        self.context.term()



import time
# from requests import get

import argparse
import parsl

# from parsl.config import Config
# from parsl.app.app import python_app
# from parsl.executors import HighThroughputExecutor
# from parsl.providers.slurm.slurm import SlurmProvider
# from parsl.launchers import SimpleLauncher


def sleep(seconds):
    import time
    time.sleep(seconds)


def double(x):
    return x * 2


def call_sleep(size):
    print("Priming ....")
    start = time.time()
    primers = [double(i) for i in range(0, 2)]
    [p.result() for p in primers]
    delta = time.time() - start
    print("Priming done in {:10.4f} s".format(delta))

    start = time.time()
    tasks = [sleep(0) for _ in range(0, size)]
    for task in tasks:
        task.result()

    delta = time.time() - start
    print("Time to complete {} tasks: {:8.3f} s".format(args.count, delta))
    print("Throughput : {:8.3f} Tasks/s".format(int(args.count) / delta))


def call_double(size, executor):
    print("Priming ....")
    start = time.time()

    primers = [executor.submit(double, i) for i in range(0, 2)]
    print("Got results : ", [p.result() for p in primers])
    delta = time.time() - start
    print("Priming done in {:10.4f} s".format(delta))

    print("Launching tasks: {}".format(size))
    start = time.time()
    tasks = [executor.submit(double, i) for i in range(0, size)]

    for task in tasks:
        task.result()

    delta = time.time() - start

    print("Time to complete {} tasks: {:8.3f} s".format(args.count, delta))
    print("Throughput : {:8.3f} Tasks/s".format(int(args.count) / delta))


def measure_latency(size, executor):
    print("Priming ....")
    start = time.time()

    primers = [executor.submit(double, i) for i in range(0, 2)]
    print("Got results : ", [p.result() for p in primers])
    delta = time.time() - start
    print("Priming done in {:10.4f} s".format(delta))

    print("Launching tasks: {}".format(size))

    start_all = time.time()
    tasks = []

    for i in range(size):
        start = time.time()
        fu = executor.submit(double, i)
        fu.result()
        delta = time.time() - start
        tasks.append(delta)

    delta_all = time.time() - start_all

    print("Time to complete {} tasks: {:8.3f} s".format(args.count, delta_all))
    print("Latency avg:{:8.3f}ms  min:{:8.3f}ms  max:{:8.3f}ms".format(
        1000 * sum(tasks) / len(tasks),
        1000 * min(tasks),
        1000 * max(tasks)))


if __name__ == '__main__':

    parser = argparse.ArgumentParser()

    parser.add_argument("-c", "--count", default="1000",
                        help="Count of apps to launch")

    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")

    args = parser.parse_args()
    parsl.set_stream_logger()
    from htex_local import config

    # from llex_local import config
    dfk = parsl.load(config)
    executor = dfk.executors["htex_local"]
    # executor = dfk.executors["llex_local"]
    # config.executors[0].worker_debug = True
    # call_double(int(args.count), executor)
    measure_latency(int(args.count), executor)



from parsl.config import Config
from parsl.channels import LocalChannel
from parsl.executors import HighThroughputExecutor
from parsl.providers import AdHocProvider

config = Config(
    executors=[
        HighThroughputExecutor(
            label='AdHoc',
            provider=AdHocProvider(
                channels=[LocalChannel(), LocalChannel()]
            )
        )
    ]
)



from parsl.providers import LocalProvider
from parsl.channels import LocalChannel
# from parsl.launchers import SimpleLauncher
from parsl.launchers import SingleNodeLauncher

from parsl.config import Config
from parsl.executors import LowLatencyExecutor

config = Config(
    executors=[
        LowLatencyExecutor(
            label="llex_local",
            # worker_debug=True,
            workers_per_node=1,
            provider=LocalProvider(
                channel=LocalChannel(),
                init_blocks=1,
                max_blocks=1,
                launcher=SingleNodeLauncher(),
            ),
        )
    ],
    strategy='none',
)



import argparse
import time
import logging
from statistics import mean, stdev

import zmq
from multiprocessing import Process, Manager

from parsl.serialize import ParslSerializer
parsl_serializer = ParslSerializer()
pack_apply_message = parsl_serializer.pack_apply_message
unpack_apply_message = parsl_serializer.unpack_apply_message
deserialize_object = parsl_serializer.deserialize

from constants import CLIENT_IP_FILE
from parsl.addresses import address_by_interface
from worker import execute_task

logger = logging.getLogger(__name__)


def simple_executor(f_all, args_all, kwargs_all, num_tasks):
    serialization_times = []
    exec_times = []
    results = []

    for i in range(num_tasks):
        task_id = i
        start_time = time.time()
        buf = pack_apply_message(f=next(f_all), args=next(args_all),
                                 kwargs=next(kwargs_all),
                                 buffer_threshold=1024 * 1024)
        serialization_times.append(time.time() - start_time)

        start_time = time.time()
        user_ns = locals()
        user_ns.update({'__builtins__': __builtins__})
        f, args, kwargs = unpack_apply_message(buf, user_ns, copy=False)
        result = execute_task(f, args, kwargs, user_ns)
        exec_times.append(time.time() - start_time)

        results.append(result)

    avg_serialization_time = sum(
        serialization_times) / len(serialization_times) * 10 ** 6
    avg_execution_time = sum(exec_times) / len(exec_times) * 10 ** 6

    return {
        "avg_serialization_time": avg_serialization_time,
        "avg_execution_time": avg_execution_time,
        "results": results
    }


def dealer_executor(f_all, args_all, kwargs_all, num_tasks, return_dict,
                    port=5559, interchange=True, warmup=10):
    label = "DEALER-INTERCHANGE-REP" if interchange else "DEALER-REP"
    logger.info("Starting executor:{}".format(label))

    serialization_times = []
    deserialization_times = []
    send_times = {}
    exec_times = {}
    results = []

    context = zmq.Context()
    dealer = context.socket(zmq.DEALER)
    dealer.bind("tcp://*:{}".format(port))

    poller = zmq.Poller()
    poller.register(dealer, zmq.POLLIN)

    num_send = 0
    num_recv = 0

    while True:
        socks = dict(poller.poll(1))
        if num_send < num_tasks:
            task_id = num_send
            task_id_bytes = task_id.to_bytes(4, "little")
            start_time = time.time()
            buf = pack_apply_message(f=next(f_all), args=next(args_all),
                                     kwargs=next(kwargs_all),
                                     buffer_threshold=1024 * 1024)
            serialization_times.append(time.time() - start_time)

            logger.debug("Manager sending task {}".format(task_id))
            send_times[task_id] = time.time()
            dealer.send_multipart([b"", task_id_bytes] + buf)
            num_send += 1

        if dealer in socks and socks[dealer] == zmq.POLLIN:
            buf = dealer.recv_multipart()
            recv_time = time.time()

            start_time = time.time()
            msg = deserialize_object(buf[2:])[0]
            deserialization_times.append(time.time() - start_time)

            logger.debug("Got message {}".format(msg))
            task_id = int.from_bytes(buf[1], "little")
            results.append(msg["result"])

            if num_recv >= warmup:
                # Ignore the first `warmup` tasks
                exec_times[task_id] = recv_time - send_times[task_id]

            num_recv += 1
            logger.debug("Dealer received result {}".format(task_id))
            if num_recv == num_tasks:
                break

    avg_serialization_time = sum(
        serialization_times) / len(serialization_times) * 10 ** 6
    avg_execution_time = sum(exec_times.values()) / len(exec_times) * 10 ** 6

    return_dict["avg_serialization_time"] = avg_serialization_time
    return_dict["avg_execution_time"] = avg_execution_time
    return_dict["results"] = results


def double(x):
    return 2 * x


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-tasks", default=10000, type=int,
                        help="Number of tasks to send for benchmark")
    parser.add_argument("--num-trials", default=10, type=int,
                        help="Number of trials to run for benchmarking")
    parser.add_argument("--warmup", default=100, type=int,
                        help="Number of warmup runs before benchmarking")
    parser.add_argument("--interchange", action="store_true", default=False,
                        help="Whether an interchange is being used")
    parser.add_argument("--client-port", default=5560, type=int,
                        help="Port for client to communicate on")
    parser.add_argument("--localhost", default=False, action="store_true",
                        help="True if communication is on localhost")
    args = parser.parse_args()

    # Write IP address to file so that workers can access it
    if not args.localhost:
        ip = address_by_interface("eth0")
        with open(CLIENT_IP_FILE, "w") as fh:
            fh.write(ip)
        print("Wrote IP address {} to file {}".format(ip, CLIENT_IP_FILE))

    # Parameters for worker requests
    def f_all():
        return (double for _ in range(args.num_tasks))

    def args_all():
        return ([i] for i in range(args.num_tasks))

    def kwargs_all():
        return ({} for _ in range(args.num_tasks))

    serialization_times = []
    execution_times = []

    # Every trial sends the same jobs again and benchmarks them
    for _ in range(args.num_trials):
        m = Manager()
        return_dict = m.dict()
        manager = Process(target=dealer_executor,
                          kwargs={"f_all": f_all(), "args_all": args_all(),
                                  "kwargs_all": kwargs_all(),
                                  "num_tasks": args.num_tasks,
                                  "port": args.client_port,
                                  "interchange": args.interchange,
                                  "warmup": args.warmup,
                                  "return_dict": return_dict})
        manager.start()
        manager.join()

        serialization_times.append(return_dict["avg_serialization_time"])
        execution_times.append(return_dict["avg_execution_time"])

    # Print stats
    label = "[DEALER-INTERCHANGE-REP]" if args.interchange else "[DEALER-REP]"
    s = stdev(serialization_times) if len(serialization_times) > 1 else 0
    print("{} Avg Serialization Time\n"
          "Mean = {:=10.4f} us, Stdev = {:=10.4f} us"
          .format(label, mean(serialization_times), s))
    s = stdev(execution_times) if len(execution_times) > 1 else 0
    print("{} Avg Execution Time\n"
          "Mean = {:=10.4f} us, Stdev = {:=10.4f} us"
          .format(label, mean(execution_times), s))



"""LowLatencyExecutor for low latency task/lambda-function execution
"""

from concurrent.futures import Future
import logging
import threading
import queue
from multiprocessing import Process, Queue

from parsl.serialize import pack_apply_message, deserialize
from parsl.executors.low_latency import zmq_pipes
from parsl.executors.low_latency import interchange
from parsl.executors.errors import ScalingFailed, DeserializationError, BadMessage, UnsupportedFeatureError
from parsl.executors.status_handling import BlockProviderExecutor
from parsl.utils import RepresentationMixin
from parsl.providers import LocalProvider

logger = logging.getLogger(__name__)


class LowLatencyExecutor(BlockProviderExecutor, RepresentationMixin):
    """
    TODO: docstring for LowLatencyExecutor
    """

    def __init__(self,
                 label='LowLatencyExecutor',
                 provider=LocalProvider(),
                 launch_cmd=None,
                 address="127.0.0.1",
                 worker_port=None,
                 worker_port_range=(54000, 55000),
                 interchange_port_range=(55000, 56000),
                 #  storage_access=None,
                 working_dir=None,
                 worker_debug=False,
                 workers_per_node=1,
                 #  cores_per_worker=1.0,
                 managed=True
                 ):
        logger.debug("Initializing LowLatencyExecutor")

        BlockProviderExecutor.__init__(self, provider=provider)
        self.label = label
        self.launch_cmd = launch_cmd
        self.provider = provider
        self.worker_debug = worker_debug
        # self.storage_access = storage_access if storage_access is not None else []
        # if len(self.storage_access) > 1:
        # raise ConfigurationError('Multiple storage access schemes are not supported')
        self.working_dir = working_dir
        self.managed = managed
        self.blocks = []
        self.workers_per_node = workers_per_node

        self._task_counter = 0
        self.address = address
        self.worker_port = worker_port
        self.worker_port_range = worker_port_range
        self.interchange_port_range = interchange_port_range
        self.run_dir = '.'

        # TODO: add debugging, logdir, other functionality to workers
        if not launch_cmd:
            self.launch_cmd = """lowlatency_worker.py -n {workers_per_node} --task_url={task_url} --logdir={logdir}"""

    def start(self):
        """Create the Interchange process and connect to it.
        """
        self.outgoing_q = zmq_pipes.TasksOutgoing(
            "127.0.0.1", self.interchange_port_range)
        self.incoming_q = zmq_pipes.ResultsIncoming(
            "127.0.0.1", self.interchange_port_range)

        self.is_alive = True

        self._queue_management_thread = None
        self._start_queue_management_thread()
        self._start_local_queue_process()

        logger.debug("Created management thread: {}"
                     .format(self._queue_management_thread))

        if self.provider:
            # debug_opts = "--debug" if self.worker_debug else ""
            l_cmd = self.launch_cmd.format(  # debug=debug_opts,
                task_url=self.worker_task_url,
                workers_per_node=self.workers_per_node,
                logdir="{}/{}".format(self.run_dir, self.label))
            self.launch_cmd = l_cmd
            logger.debug("Launch command: {}".format(self.launch_cmd))

            self._scaling_enabled = True
            logger.debug(
                "Starting LowLatencyExecutor with provider:\n%s", self.provider)
            if hasattr(self.provider, 'init_blocks'):
                try:
                    for i in range(self.provider.init_blocks):
                        block = self.provider.submit(
                            self.launch_cmd, self.workers_per_node)
                        logger.debug("Launched block {}:{}".format(i, block))
                        if not block:
                            raise(ScalingFailed(self,
                                                "Attempts to provision nodes via provider has failed"))
                        self.blocks.extend([block])

                except Exception as e:
                    logger.error("Scaling out failed: {}".format(e))
                    raise e
        else:
            self._scaling_enabled = False
            logger.debug("Starting LowLatencyExecutor with no provider")

    def _start_local_queue_process(self):
        """ TODO: docstring """

        comm_q = Queue(maxsize=10)
        self.queue_proc = Process(target=interchange.starter,
                                  args=(comm_q,),
                                  kwargs={"client_ports": (self.outgoing_q.port,
                                                           self.incoming_q.port),
                                          "worker_port": self.worker_port,
                                          "worker_port_range": self.worker_port_range
                                          # TODO: logdir and logging level
                                          })
        self.queue_proc.start()

        try:
            worker_port = comm_q.get(block=True, timeout=120)
            logger.debug(
                "Got worker port {} from interchange".format(worker_port))
        except queue.Empty:
            logger.error(
                "Interchange has not completed initialization in 120s. Aborting")
            raise Exception("Interchange failed to start")

        self.worker_task_url = "tcp://{}:{}".format(
            self.address, worker_port)

    def _start_queue_management_thread(self):
        """ TODO: docstring """
        if self._queue_management_thread is None:
            logger.debug("Starting queue management thread")
            self._queue_management_thread = threading.Thread(
                target=self._queue_management_worker)
            self._queue_management_thread.daemon = True
            self._queue_management_thread.start()
            logger.debug("Started queue management thread")

        else:
            logger.debug("Management thread already exists, returning")

    def _queue_management_worker(self):
        """ TODO: docstring """
        logger.debug("[MTHREAD] queue management worker starting")

        while not self.bad_state_is_set:
            task_id, buf = self.incoming_q.get()  # TODO: why does this hang?
            msg = deserialize(buf)[0]
            # TODO: handle exceptions
            task_fut = self.tasks[task_id]
            logger.debug("Got response for task id {}".format(task_id))

            if "result" in msg:
                task_fut.set_result(msg["result"])

            elif "exception" in msg:
                # TODO: handle exception
                pass
            elif 'exception' in msg:
                logger.warning("Task: {} has returned with an exception")
                try:
                    s = deserialize(msg['exception'])
                    exception = ValueError("Remote exception description: {}".format(s))
                    task_fut.set_exception(exception)
                except Exception as e:
                    # TODO could be a proper wrapped exception?
                    task_fut.set_exception(
                        DeserializationError("Received exception, but handling also threw an exception: {}".format(e)))

            else:
                raise BadMessage(
                    "Message received is neither result nor exception")

            if not self.is_alive:
                break

        logger.info("[MTHREAD] queue management worker finished")

    def submit(self, func, resource_specification, *args, **kwargs):
        """ TODO: docstring """
        if resource_specification:
            logger.error("Ignoring the resource specification. "
                         "Parsl resource specification is not supported in LowLatency Executor. "
                         "Please check WorkQueueExecutor if resource specification is needed.")
            raise UnsupportedFeatureError('resource specification', 'LowLatency Executor', 'WorkQueue Executor')

        if self.bad_state_is_set:
            raise self.executor_exception

        self._task_counter += 1
        task_id = self._task_counter

        logger.debug(
            "Pushing function {} to queue with args {}".format(func, args))

        self.tasks[task_id] = Future()

        fn_buf = pack_apply_message(func, args, kwargs,
                                    buffer_threshold=1024 * 1024)

        # Post task to the the outgoing queue
        self.outgoing_q.put(task_id, fn_buf)

        # Return the future
        return self.tasks[task_id]

    @property
    def scaling_enabled(self):
        return self._scaling_enabled

    def scale_out(self, blocks=1):
        """Scales out the number of active workers by the number of blocks specified.

        Parameters
        ----------

        blocks : int
             # of blocks to scale out. Default=1
        """
        r = []
        for i in range(blocks):
            if self.provider:
                try:
                    block = self.provider.submit(
                        self.launch_cmd, self.workers_per_node)
                    logger.debug("Launched block {}:{}".format(i, block))
                    # TODO: use exceptions for this
                    if not block:
                        self._fail_job_async(None, "Failed to launch block")
                    self.blocks.extend([block])
                except Exception as ex:
                    self._fail_job_async(None, "Failed to launch block: {}".format(ex))
            else:
                logger.error("No execution provider available")
                r = None
        return r

    def scale_in(self, blocks):
        """Scale in the number of active blocks by specified amount.

        The scale in method here is very rude. It doesn't give the workers
        the opportunity to finish current tasks or cleanup. This is tracked
        in issue #530
        """
        to_kill = self.blocks[:blocks]
        if self.provider:
            r = self.provider.cancel(to_kill)
        return self._filter_scale_in_ids(to_kill, r)

    def _get_job_ids(self):
        return self.blocks

    def shutdown(self, hub=True, targets='all', block=False):
        """Shutdown the executor, including all workers and controllers.

        This is not implemented.

        Kwargs:
            - hub (Bool): Whether the hub should be shutdown, Default:True,
            - targets (list of ints| 'all'): List of block id's to kill, Default:'all'
            - block (Bool): To block for confirmations or not
        """

        logger.warning("Attempting LowLatencyExecutor shutdown")
        # self.outgoing_q.close()
        # self.incoming_q.close()
        self.queue_proc.terminate()
        logger.warning("Finished LowLatencyExecutor shutdown attempt")
        return True



#!/usr/bin/env python3

import logging
import zmq
# import time


class Interchange(object):
    """ TODO: docstring """

    def __init__(self,
                 client_address="127.0.0.1",
                 client_ports=(50055, 50056),
                 worker_port=None,
                 worker_port_range=(54000, 55000)
                 ):
        global logger
        start_file_logger("interchange.log")
        logger.info("Init Interchange")

        self.context = zmq.Context()
        self.task_incoming = self.context.socket(zmq.ROUTER)
        self.result_outgoing = self.context.socket(zmq.DEALER)
        self.worker_messages = self.context.socket(zmq.DEALER)

        self.result_outgoing.set_hwm(0)

        task_address = "tcp://{}:{}".format(client_address, client_ports[0])
        result_address = "tcp://{}:{}".format(client_address, client_ports[1])
        self.task_incoming.connect(task_address)
        self.result_outgoing.connect(result_address)

        logger.debug("Client task address: {}".format(task_address))
        logger.debug("Client result address: {}".format(result_address))

        self.worker_port = worker_port
        self.worker_port_range = worker_port_range

        if self.worker_port:
            worker_task_address = "tcp://*:{}".format(self.worker_port)
            self.worker_messages.bind(worker_task_address)
            logger.debug("Worker task address: {}".format(worker_task_address))

        else:
            self.worker_port = self.worker_messages.bind_to_random_port(
                'tcp://*',
                min_port=worker_port_range[0],
                max_port=worker_port_range[1], max_tries=100)

            logger.debug(
                "Worker task address: tcp://*:{}".format(self.worker_port))

        self.poller = zmq.Poller()
        self.poller.register(self.task_incoming, zmq.POLLIN)
        self.poller.register(self.worker_messages, zmq.POLLIN)
        logger.debug("Init complete")

    def start(self):
        """ TODO: docstring """
        logger.info("Starting interchange")
        # last = time.time()

        while True:
            # active_flag = False
            socks = dict(self.poller.poll(1))

            if socks.get(self.task_incoming) == zmq.POLLIN:
                message = self.task_incoming.recv_multipart()
                logger.debug("Got new task from client")
                self.worker_messages.send_multipart(message)
                logger.debug("Sent task to worker")
                # active_flag = True
                # last = time.time()

            if socks.get(self.worker_messages) == zmq.POLLIN:
                message = self.worker_messages.recv_multipart()
                logger.debug("Got new result from worker")
                # self.result_outgoing.send_multipart(message)
                self.result_outgoing.send_multipart(message[1:])

                logger.debug("Sent result to client")
                # active_flag = True
                # last = time.time()

            # if not active_flag and last + 1 < time.time():
            #    logger.debug("Nothing in the past 1s round")
            #    last = time.time()


def start_file_logger(filename, name='interchange', level=logging.DEBUG, format_string=None):
    """Add a stream log handler.

    Parameters
    ---------

    filename: string
        Name of the file to write logs to. Required.
    name: string
        Logger name. Default="parsl.executors.interchange"
    level: logging.LEVEL
        Set the logging level. Default=logging.DEBUG
        - format_string (string): Set the format string
    format_string: string
        Format string to use.

    Returns
    -------
        None.
    """
    if format_string is None:
        format_string = "%(asctime)s %(name)s:%(lineno)d [%(levelname)s]  %(message)s"

    global logger
    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.FileHandler(filename)
    handler.setLevel(level)
    formatter = logging.Formatter(format_string, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def starter(comm_q, *args, **kwargs):
    """Start the interchange process

    The executor is expected to call this function. The args, kwargs match that of the Interchange.__init__
    """
    # logger = multiprocessing.get_logger()
    ic = Interchange(*args, **kwargs)
    comm_q.put(ic.worker_port)
    ic.start()
    logger.debug("Port information sent back to client")



import argparse
import logging

import zmq
from multiprocessing import Process

from parsl.serialize import ParslSerializer
parsl_serializer = ParslSerializer()
unpack_apply_message = parsl_serializer.unpack_apply_message
serialize_object = parsl_serializer.serialize

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")

from constants import CLIENT_IP_FILE, INTERCHANGE_IP_FILE
from utils import ping_time


def execute_task(f, args, kwargs, user_ns):
    """
    Deserialize the buffer and execute the task.

    Returns the result or exception.
    """
    fname = getattr(f, '__name__', 'f')
    prefix = "parsl_"
    fname = prefix + "f"
    argname = prefix + "args"
    kwargname = prefix + "kwargs"
    resultname = prefix + "result"

    user_ns.update({fname: f,
                    argname: args,
                    kwargname: kwargs,
                    resultname: resultname})

    code = "{0} = {1}(*{2}, **{3})".format(resultname, fname,
                                           argname, kwargname)
    try:
        exec(code, user_ns, user_ns)

    except Exception as e:
        logger.warning("Caught exception; will raise it: {}".format(e))
        raise e

    else:
        return user_ns.get(resultname)


def dealer_worker(worker_id, ip="localhost", port=5560):
    context = zmq.Context()
    socket = context.socket(zmq.REP)
    socket.connect("tcp://{}:{}".format(ip, port))
    print("Starting worker {}".format(worker_id))

    task_ids_received = []

    while True:
        bufs = socket.recv_multipart()
        task_id = int.from_bytes(bufs[0], "little")
        task_ids_received.append(task_id)

        user_ns = locals()
        user_ns.update({'__builtins__': __builtins__})
        f, args, kwargs = unpack_apply_message(bufs[1:], user_ns, copy=False)

        logger.debug("Worker {} received task {}".format(worker_id, task_id))
        result = execute_task(f, args, kwargs, user_ns)
        logger.debug("Worker result: {}".format(result))
        reply = {"result": result, "worker_id": worker_id}
        socket.send_multipart([bufs[0]] + serialize_object(reply))

        print("Worker {} received {} tasks"
              .format(worker_id, len(task_ids_received)))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-workers", default=1, type=int,
                        help="Number of workers to use for Dealer-Rep")
    parser.add_argument("--localhost", default=False, action="store_true",
                        help="True if communication is on localhost")
    parser.add_argument("--worker-port", default=5559, type=int,
                        help="Port for workers to communicate on")
    parser.add_argument("--interchange", action="store_true", default=False,
                        help="Whether an interchange is being used")
    args = parser.parse_args()

    # Read in IP address to communicate to
    if not args.localhost:
        ip_file = INTERCHANGE_IP_FILE if args.interchange else CLIENT_IP_FILE
        with open(ip_file, "r") as fh:
            ip = fh.read().strip()
        print("Read IP {} from file {}".format(ip, ip_file))
        print("Ping time to IP {}: {} us".format(ip, ping_time(ip)))
    else:
        ip = "localhost"

    workers = []
    for i in range(args.num_workers):
        worker = Process(target=dealer_worker,
                         kwargs={"worker_id": i, "ip": ip,
                                 "port": args.worker_port})
        worker.daemon = True
        worker.start()
        workers.append(worker)

    for worker in workers:
        worker.join()



import logging
import argparse

from multiprocessing import Process
import zmq

from constants import CLIENT_IP_FILE, INTERCHANGE_IP_FILE
from parsl.addresses import address_by_interface

logger = logging.getLogger(__name__)


def dealer_interchange(manager_ip="localhost", manager_port=5559,
                       worker_port=5560):
    context = zmq.Context()
    incoming = context.socket(zmq.ROUTER)
    outgoing = context.socket(zmq.DEALER)

    incoming.connect("tcp://{}:{}".format(manager_ip, manager_port))
    outgoing.bind("tcp://*:{}".format(worker_port))

    poller = zmq.Poller()
    poller.register(incoming, zmq.POLLIN)
    poller.register(outgoing, zmq.POLLIN)

    while True:
        socks = dict(poller.poll(1))

        if socks.get(incoming) == zmq.POLLIN:
            message = incoming.recv_multipart()
            logger.debug("[interchange] New task {}".format(message))
            outgoing.send_multipart(message)

        if socks.get(outgoing) == zmq.POLLIN:
            message = outgoing.recv_multipart()
            logger.debug("[interchange] New result {}".format(message))
            incoming.send_multipart(message)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--worker-port", default=5559, type=int,
                        help="Port for workers to communicate on")
    parser.add_argument("--client-port", default=5560, type=int,
                        help="Port for client to communicate on")
    parser.add_argument("--localhost", default=False, action="store_true",
                        help="True if communication is on localhost")
    args = parser.parse_args()

    if not args.localhost:
        with open(CLIENT_IP_FILE, "r") as fh:
            client_ip = fh.read().strip()
        print("Read IP {} from file {}".format(client_ip, CLIENT_IP_FILE))

        interchange_ip = address_by_interface("eth0")
        with open(INTERCHANGE_IP_FILE, "w") as fh:
            fh.write(interchange_ip)
        print("Wrote IP address {} to file {}"
              .format(interchange_ip, INTERCHANGE_IP_FILE))
    else:
        client_ip = "localhost"

    interchange = Process(target=dealer_interchange,
                          kwargs={"manager_ip": client_ip,
                                  "manager_port": args.client_port,
                                  "worker_port": args.worker_port})
    interchange.daemon = True
    interchange.start()
    interchange.join()



import parsl
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.executors.threads import ThreadPoolExecutor
from parsl.providers import LocalProvider
from parsl.app.app import python_app
from parsl.launchers import SrunLauncher
from parsl.providers.slurm.slurm import SlurmProvider


config = Config(
    executors=[
        HighThroughputExecutor(
            label='htex_local',
            cores_per_worker=1,
            max_workers=5,
            provider=LocalProvider(
                init_blocks=1,
                max_blocks=1,
            ),
        )
    ],
    strategy='none',
)


def local_setup():
    global dfk
    dfk = parsl.load(config)


def local_teardown():
    parsl.clear()


@python_app(executors=['htex_local'])
def sleeper(dur=1):
    import time
    time.sleep(dur)


@python_app(executors=['threads'])
def cpu_stress(dur=2):
    import time
    s = 0
    start = time.time()
    for i in range(10**8):
        s += i
        if time.time() - start >= dur:
            break
    return s


@python_app(executors=['midway_htex'])
def add(n):
    s = 0
    for i in range(n):
        s += i
    return s


if __name__ == "__main__":

    tasks = [sleeper() for i in range(5)]
    results = [i.result() for i in tasks]
    print("Done with initial test. The results are ", results)

    thread_executors = [ThreadPoolExecutor(
        label='threads',
        max_threads=4)
    ]
    dfk.add_executors(executors=thread_executors)
    tasks = [cpu_stress() for i in range(10)]
    results = [i.result() for i in tasks]
    print("Successfully added thread executor and ran with it. The results are ", results)

    htex_executors = [
        HighThroughputExecutor(
            label="midway_htex",
            # worker_debug=True,
            cores_per_worker=1,
            provider=SlurmProvider(
                'broadwl',
                launcher=SrunLauncher(),
                scheduler_options='#SBATCH --exclusive',
                worker_init='source activate dask',
                init_blocks=1,
                max_blocks=1,
                min_blocks=1,
                nodes_per_block=2,
                walltime='00:10:00',
            ),
        )
    ]
    dfk.add_executors(executors=htex_executors)
    tasks = [add(i) for i in range(60)]
    results = [i.result() for i in tasks]
    print("Successfully added midway_htex executor and ran with it. The results are ", results)
    print("Done testing")



import argparse
import time
import parsl
import pytest

from parsl.providers import LocalProvider
from parsl.channels import LocalChannel
from parsl.launchers import SingleNodeLauncher
from parsl.config import Config
from parsl.executors import HighThroughputExecutor

try:
    from parsl.addresses import address_by_interface
    address = address_by_interface('enx00e112002c62')
except OSError:
    print("***************ERROR****************************")
    print("Interface doesn't exist, falling back to default, 127.0.0.1")
    print("***************ERROR****************************")
    address = '127.0.01'


config = Config(
    executors=[
        HighThroughputExecutor(
            poll_period=1,
            address=address,
            label="htex_local",
            worker_debug=True,
            cores_per_worker=1,
            heartbeat_period=1,
            heartbeat_threshold=2,
            worker_ports=(54759, 54330),
            suppress_failure=True,
            provider=LocalProvider(
                channel=LocalChannel(),
                init_blocks=0,
                max_blocks=0,
                # tasks_per_node=1,  # For HighThroughputExecutor, this option should in most cases be 1
                launcher=SingleNodeLauncher(),
            ),
        )
    ],
    strategy='none',
)


# config.executors[0].provider.tasks_per_node = 4
from parsl.app.app import python_app  # , bash_app


@python_app
def double(x):
    return x * 2


@python_app
def platform(sleep=10, stdout=None):
    import platform
    import time
    time.sleep(sleep)
    return platform.uname()


@pytest.mark.skip('not asserting anything')
def test_simple(n=2, dur=10):
    """ Tests whether the right exception is returned when a manager is lost.
    Check whether the error string reports manager loss on node X.

    """
    start = time.time()

    # Prime a worker
    double(5).result()

    # Launch the slow task to a failing node.
    x = platform(sleep=dur)

    time.sleep(0.1)

    dfk = parsl.dfk()
    job_ids = dfk.executors['htex_local'].provider.resources.keys()
    dfk.executors['htex_local'].provider.cancel(job_ids)
    print("Result : ", x.result())
    print("Duration : {0}s".format(time.time() - start))
    print("[TEST STATUS] test_parallel_for [SUCCESS]")
    return True


@pytest.mark.skip('not asserting anything')
def test_manager_fail(n=2, dur=10):
    """ Test manager failure due to intermittent n/w loss.
    Run parsl on laptop, start the worker on a remote node.
    Once connected and task is executed, kill the network
    confirm that the interchange does not error out.
    """
    start = time.time()

    # Prime a worker
    double(5).result()

    # Launch the slow task to a failing node.
    x = platform(sleep=dur)
    x.result()

    # At this point we know that the worker is connected.
    # Now we need to kill the network to mimic a n/w failure.
    time.sleep(0.1)
    print('*' * 80)
    print("Manually kill the network to create a n/w failure")
    print("We should see suppression of the failure")
    print('*' * 80)

    print("Duration : {0}s".format(time.time() - start))
    print("[TEST STATUS] test_parallel_for [SUCCESS]")
    dfk = parsl.dfk()
    for i in range(200):
        print(dfk.executors['htex_local'].connected_workers)
        time.sleep(1)

    return True


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--sleep", default="4")
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    parsl.load(config)

    # x = test_simple(int(args.count))
    # x = test_imports()
    # x = test_simple(int(args.count), int(args.sleep))
    x = test_manager_fail(int(args.count), int(args.sleep))



import subprocess


def ping_time(ip, n=5):
    """
    Returns the average ping time in microseconds.

    Note: This function is inherently platform specific.
    It currently works on Midway.
    """
    cmd = "ping {} -c {}".format(ip, n)
    p = subprocess.Popen(cmd.split(" "), stdout=subprocess.PIPE)
    output = str(p.communicate()[0])
    stats = output.split("\n")[-1].split(" = ")[-1].split("/")
    avg_ping_time = float(stats[1])  # In ms
    return avg_ping_time * 1000



import pytest

from parsl import python_app
from parsl.tests.configs.htex_local import fresh_config
from parsl.executors.errors import SerializationError

local_config = fresh_config()
local_config.retries = 2


@python_app
def fail_pickling(x):
    return next(x)


def generator():
    num = 0
    while True:
        yield num
        num += 1


@pytest.mark.local
def test_serialization_error():

    gen = generator()
    x = fail_pickling(gen)

    try:
        x.result()
    except Exception as e:
        assert isinstance(e, SerializationError)
    else:
        raise ValueError



#!/usr/bin/env python3

"""Test process layer.

<bash>
Foreach ….
    10 K app calls in parallel

<bash 30 s> wait on all for loop -> .csv

Foreach .. from .csv
    100K app calls in parallel. Needs softImage

<Bash> wait on all for loop

"""

import parsl
from parsl import bash_app, python_app, DataFlowKernel, ThreadPoolExecutor
import os
import shutil
import random
import argparse


workers = ThreadPoolExecutor(max_workers=8)
dfk = DataFlowKernel(executors=[workers])


def create_dirs(cwd):

    for dir in ['relax.01', 'relax.02', 'relax.03']:
        rel_dir = '{0}/{1}'.format(cwd, dir)
        if os.path.exists(rel_dir):
            shutil.rmtree(rel_dir)
        os.makedirs(rel_dir)
        for i in range(0, random.randint(1, 5)):
            rdir = '{0}/{1}'.format(rel_dir, i)
            os.makedirs(rdir)
            with open('{0}/results'.format(rdir), 'w') as f:
                f.write("{0} {1} - test data\n".format(i, dir))

    for dir in ['neb01', 'neb02', 'neb03', 'neb04']:
        rel_dir = '{0}/{1}'.format(cwd, dir)
        if os.path.exists(rel_dir):
            shutil.rmtree(rel_dir)
        os.makedirs(rel_dir)
        with open('{0}/{1}.txt'.format(rel_dir, dir), 'w') as f:
            f.write("{0} test data\n".format(rel_dir))


@python_app(data_flow_kernel=dfk)
def ls(pwd, outputs=[]):
    import os
    items = os.listdir(pwd)
    with open(outputs[0], 'w') as f:
        f.write(' '.join(items))
        f.write('\n')
    # Returning list of items in current dir as python object
    return items


@bash_app(data_flow_kernel=dfk)
def catter(dir, dur, inputs=[], outputs=[], stdout=None, stderr=None):
    cmd_line = 'cd {0}; echo "sleeping... "; sleep {1}'
    return cmd_line


@bash_app(data_flow_kernel=dfk)
def light_app(dir, dur, inputs=[], outputs=[], stdout=None, stderr=None):
    cmd_line = 'cd {0}; echo "light_app" > {outputs[0]} ; sleep {1}'
    return cmd_line


@bash_app(data_flow_kernel=dfk)
def csv_maker(dir, count, dur, inputs=[], outputs=[], stdout=None, stderr=None):

    cmd_line = """cd {0};
    # create a file with count lines
    shuf -i 1-{1} &> {outputs[0]}
    sleep {2}
    """
    return cmd_line


def main(count):

    # <Bash app >
    c1 = catter('.', 0, stdout='outputs/catter1.out',
                stderr='outputs/catter1.err')

    light_loop = []
    # Foreach parallel loop 10K calls
    for i in range(count):
        if i % 1000:
            print("Launching light : ", i)
        outname = 'outputs/light{0}'.format(i)
        loop1 = light_app('.', 0, inputs=[c1], outputs=[
                          outname + '.txt'], stdout=outname + '.out')
        light_loop.extend([loop1])

    # <Bash app dependent on for loop>
    c2 = csv_maker('.', count * 10, 0, inputs=light_loop, outputs=['csv_maker.csv'],
                   stdout='outputs/catter2.out', stderr='outputs/catter2.err')
    csv_file = c2.outputs[0]

    # This is a blocking call that forces the workflow to wait for the csv_file to be produced.
    lines = open(csv_file.result(), 'r').readlines()
    mid_loop = []

    # Foreach parallel loop 10K calls
    for i in lines:
        i = i.strip()
        outname = 'outputs/mid{0}'.format(i)
        loop1 = light_app('.', 0, inputs=[c1], outputs=[
                          outname + '.txt'], stdout=outname + '.out')
        mid_loop.extend([loop1])

    # <Bash app dependent on mid for loop>
    c3 = catter('.', 3, inputs=mid_loop,
                stdout='outputs/catter3.out', stderr='outputs/catter3.err')
    return c3


def test_HEDM(count=10):
    x = main(count)
    x.result()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--count", default="10",
                        help="Count of apps to launch")
    parser.add_argument("-d", "--debug", action='store_true',
                        help="Count of apps to launch")
    args = parser.parse_args()

    if args.debug:
        parsl.set_stream_logger()

    # Remove outputs and recreate
    shutil.rmtree("outputs", True)
    os.mkdir("outputs")
    # x = test_parallel_for(int(args.count))
    x = main(int(args.count))
    print("Waiting for workflow result")
    x.result()



#!/usr/bin/env python3

import argparse
import logging
import os
import uuid
# import zmq
from multiprocessing import Process

from parsl.serialize import unpack_apply_message, serialize
from parsl.executors.low_latency import zmq_pipes

logger = logging.getLogger(__name__)


def execute_task(f, args, kwargs, user_ns):
    """
    Deserialize the buffer and execute the task.

    # Returns the result or exception.
    """
    fname = getattr(f, '__name__', 'f')
    prefix = "parsl_"
    fname = prefix + "f"
    argname = prefix + "args"
    kwargname = prefix + "kwargs"
    resultname = prefix + "result"

    user_ns.update({fname: f,
                    argname: args,
                    kwargname: kwargs,
                    resultname: resultname})

    code = "{0} = {1}(*{2}, **{3})".format(resultname, fname,
                                           argname, kwargname)
    try:
        exec(code, user_ns, user_ns)

    except Exception as e:
        logger.warning("Caught exception; will raise it: {}".format(e))
        raise e

    else:
        return user_ns.get(resultname)


def start_file_logger(filename, rank, name='parsl', level=logging.DEBUG, format_string=None):
    """Add a stream log handler.

    Args:
        - filename (string): Name of the file to write logs to
        - name (string): Logger name
        - level (logging.LEVEL): Set the logging level.
        - format_string (string): Set the format string

    Returns:
       -  None
    """

    try:
        os.makedirs(os.path.dirname(filename), 511, True)
    except Exception as e:
        print("Caught exception with trying to make log dirs: {}".format(e))

    if format_string is None:
        format_string = "%(asctime)s %(name)s:%(lineno)d Rank:{0} [%(levelname)s]  %(message)s".format(
            rank)
    global logger
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(filename)
    handler.setLevel(level)
    formatter = logging.Formatter(format_string, datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def worker(worker_id, task_url, debug=True, logdir="workers", uid="1"):
    """ TODO: docstring

    TODO : Cleanup debug, logdir and uid to function correctly
    """

    start_file_logger('{}/{}/worker_{}.log'.format(logdir, uid, worker_id),
                      0,
                      level=logging.DEBUG if debug is True else logging.INFO)

    logger.info("Starting worker {}".format(worker_id))

    task_ids_received = []

    message_q = zmq_pipes.WorkerMessages(task_url)

    while True:
        print("Worker loop iteration starting")
        task_id, buf = message_q.get()
        task_ids_received.append(task_id)

        user_ns = locals()
        user_ns.update({'__builtins__': __builtins__})
        f, args, kwargs = unpack_apply_message(buf, user_ns, copy=False)

        logger.debug("Worker {} received task {}".format(worker_id, task_id))
        result = execute_task(f, args, kwargs, user_ns)
        logger.debug("Worker {} completed task {}".format(worker_id, task_id))

        reply = {"result": result, "worker_id": worker_id}
        message_q.put(task_id, serialize(reply))
        logger.debug("Result sent")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", "--workers_per_node", default=1, type=int,
                        help="Number of workers to kick off. Default=1")
    parser.add_argument("-l", "--logdir", default="lowlatency_worker_logs",
                        help="LowLatency worker log directory")
    parser.add_argument("-t", "--task_url", required=True,
                        help="REQUIRED: ZMQ url for receiving tasks")
    parser.add_argument("-u", "--uid", default=str(uuid.uuid4()).split('-')[-1],
                        help="Unique identifier string for Manager")

    args = parser.parse_args()

    workers = []
    for i in range(args.workers_per_node):
        worker = Process(target=worker,
                         kwargs={"worker_id": i,
                                 "task_url": args.task_url,
                                 "logdir": args.logdir,
                                 "uid": args.uid
                                 })
        worker.daemon = True
        worker.start()
        workers.append(worker)

    for worker in workers:
        worker.join()



import time

import pytest

import parsl
from parsl.app.app import python_app
from parsl.tests.configs.htex_local import fresh_config


def local_setup():
    global dfk
    config = fresh_config()
    config.executors[0].init_blocks = 0
    config.executors[0].min_blocks = 0
    config.executors[0].max_blocks = 4
    dfk = parsl.load(config)


def local_teardown():
    parsl.dfk().cleanup()
    parsl.clear()


@python_app
def diamond(sleep=0, inputs=[]):
    import time
    time.sleep(sleep)
    return sum(inputs)


@pytest.mark.local
@pytest.mark.skip('slow and does not assert anything')
def test_python(width=10):
    """Diamond pattern to scale from 0 -> 1 -> N -> 1 -> 0 """

    stage_1 = [diamond(sleep=60, inputs=[0])]

    stage_2 = []
    for i in range(0, width):
        stage_2.extend([diamond(sleep=20, inputs=stage_1)])

    stage_3 = [diamond(sleep=30, inputs=stage_2)]

    if not stage_3[0].done():
        time.sleep(30)
        for sitename in dfk.executors:
            print(dfk.executors[sitename].status())


if __name__ == "__main__":
    test_python()
